package edu.psu.geovista.datamining.data.HD;

/* --------------------------- Package ---------------------------- */
import java.awt.*;
import java.util.*;
import edu.psu.geovista.datamining.util.*;
//import edu.psu.geovista.datamining.ordering.*;

/**
 * <p>Title: Interactive Spatial Data Mining for Large and High-Dimensional Data</p>
 * <p>Description: </p>
 * <p>Copyright: Copyright (c) 2003</p>
 * <p>Company: GeoVISTA Center and Department of Geography, Pennsylvania State University</p>
 * @author Diansheng Guo
 * @version 1.0
 */
public class BasicCell
    implements Comparable/*, GenericOrderableInstance*/ {
  //This field is used to sort points according to a KEY. This key should be
  //assigned before the sorting.
  private double key = 0;

  //Unique id
  private int ID;

  //Dimension IDs (0, 1, ..) for defining this cell.
  private int dims[];

  //Dimension ranges the defines this cells. They can be null if the cell is
  //defined in other ways, as for SOM methods.
  private double ranges[][];
  private double syntheticValue[];

  //////////////////////////////////////////////////////////////////
  //Data variables
  private boolean keepinstances = false;
  private int inst_count = 0;
  private BasicInstanceInterface[] instances = null;
  private double sums[]; //Summary statistics ...
  private double maxs[];
  private double mins[];

  //Here is a flexible reference to any other info object created by application
  //class that uses SpatialPoint but wants to add more parameters to the point.
  private Object infoObject = null;
  private Object peer = null;

  //////////////////////////////////////////////////////////////////
  //Computational variables
  private boolean dense = true;
  private int ordering; //The ordering, encoding, and special properties
  private double encoding;
  private boolean special = false;
  private int majorClusterId = -1; //The cluster ID
  private int minorClusterId = -1;

  //////////////////////////////////////////////////////////////////
  //Visual variables
  public Color color = ColorSetting.disabled; //Color
  public boolean selected = false; //Selection
  public boolean showSpecial = false;

  /***
   * NUll Constructor: the configure() method must be called before adding instances.
   */
  public BasicCell() {}

  public BasicCell(int id,
                   int[] dimensions, //dimension id, e.g., 0, 1, ...
                   double[][] dimranges, //can be null
                   boolean keepinstances) {
    this.configure(id, dimensions, dimranges, keepinstances);
  }

  /*****
   * Configurer method.
   * **/
  public void configure(int id,
                        int[] dimensions, //dimension id, e.g., 0, 1, ...
                        double[][] dimranges, //can be null
                        boolean keepinstances) {
    this.ID = id;
    this.syntheticValue = new double[dimensions.length];
    this.dims = new int[dimensions.length];
    System.arraycopy(dimensions, 0, this.dims, 0, this.dims.length);

    if (dimranges != null) {
      if (dimranges.length != this.dims.length) {
        System.out.println(
            "ranges.length != dimensions.length in BasicCell() constructor");
        return;
      }
      if (dimranges[0].length != 2) {
        System.out.println("each range must have only 2 values: min and max");
        return;
      }

      this.ranges = new double[this.dims.length][2];
      for (int i = 0; i < this.dims.length; i++) {
        this.ranges[i][0] = (dimranges[i][0] < dimranges[i][1]) ?
            dimranges[i][0] :
            dimranges[i][1];

        this.ranges[i][1] = (dimranges[i][0] > dimranges[i][1]) ?
            dimranges[i][0] :
            dimranges[i][1];
      }
    }
    //Summary statistics
    sums = new double[dimensions.length];
    maxs = new double[dimensions.length];
    mins = new double[dimensions.length];
    for (int i = 0; i < dimensions.length; i++) {
      sums[i] = 0;
      maxs[i] = Integer.MIN_VALUE;
      mins[i] = Integer.MAX_VALUE;
    }

    this.keepinstances = keepinstances;
    this.inst_count = 0;

    if (keepinstances) {
      this.instances = new BasicInstanceInterface[10];
    }
  }

  /**Should be overwritten by subclasses. */
  public int[] getInstanceIndice() {
    int[] temp = new int[this.inst_count];
    for (int i = 0; i < this.inst_count; i++)
      temp[i] = this.instances[i].getId();
    return temp;
  }

  public double getAverage(int dim_idx) {
    return this.sums[dim_idx] / (double) (this.inst_count);
  }

  public double getDataMax(int dim_idx) {
    return this.maxs[dim_idx];
  }

  public double getDataMin(int dim_idx) {
    return this.mins[dim_idx];
  }

  public double getSum(int dim_idx) {
    return this.sums[dim_idx];
  }

  /** Add a new point to this cell
   *  If the level is less than the required level, one point per cell.
   *  if the level is equal to or larger than the required level,
   *  max_point_in_one_cell points per cell.
   *
   * If the above limtes broken, generate new children cells and pass points to
   * children.
   *
   *
   *  */
  public void addInstance(BasicInstanceInterface instance) {
    //check if the point is within the range of this cell
    int dim = -1;
    double value;
    for (int i = 0; i < dims.length; i++) {
      dim = this.dims[i];
      //we're getting the original value instead of normalized value.
      value = instance.getOriginalElementValue(dim);
      this.sums[i] = this.sums[i] + value;
      this.setSyntheticValue(i, this.sums[i] / (inst_count + 1));
      if (value > this.maxs[i])this.maxs[i] = value;
      if (value < this.mins[i])this.mins[i] = value;
    }

    if (!keepinstances) {
      inst_count++;
      return;
    }

    if (inst_count < instances.length)
      instances[inst_count] = instance;
    else {
      BasicInstanceInterface[] temp = new BasicInstanceInterface[this.instances.length * 2];
      System.arraycopy(this.instances, 0, temp, 0, this.instances.length);
      this.instances = temp;
      this.instances[inst_count] = instance;
    }
    inst_count++;
  }

  public double getSyntheticValue(int dim) {
    return this.syntheticValue[dim];
  }
//  public double[] getSyntheticValue() {
//    return this.syntheticValue;
//  }
//  public void setSyntheticValue(double synthetic[]) {
//    this.syntheticValue = synthetic;
//  }

  public void setSyntheticValue(int dim, double val){
    this.syntheticValue[dim] = val;
  }

  public void setPeer(Object peer) {
    this.peer = peer;
  }

  public Object getPeer() {
    return this.peer;
  }

  public int getCountOfInstances() {
    return this.inst_count;
  }

  public int getDimensionality() {
    return this.dims.length;
  }

  public int[] getDimensions() {
    return this.dims;
  }

  public double[][] getRanges() {
    return this.ranges;
  }

  public boolean equalTo(BasicCell another) {
    if (another.ID == this.ID)
      return true;
    else
      return false;
  }

  public BasicInstanceInterface[] getInstances() {
    return this.instances;
  }

  public void setDense(boolean dense) {
    this.dense = dense;
  }

  public boolean isDense() {
    return this.dense;
  }

  public void setHDColor(Color c) {
    this.color = c;
  }

  public void setSelected(boolean selected) {
    this.selected = selected;
  }

  /**
   * The class that uses the info object can instantiate an object of a specific
   * class and set it as an info object. After getting it, it should be casted to
   * whatever class it actually is.
   * */
  public Object getInfoObject() {
    return this.infoObject;
  }

  public void setInfoObject(Object info) {
    this.infoObject = info;
  }

  /**
   *  Methods for geting the ordering, encoding, and cluster ID
   * */
  public int getOrdering() {
    return this.ordering;
  }

  public double getEncoding() {
    return this.encoding;
  }

  public boolean isOutlier() {
    return this.special;
  }

  public int getMajorClusterId() {
    return this.majorClusterId;
  }

  public int getMinorClusterId() {
    return this.minorClusterId;
  }

  /* Methods for seting the ordering, encoding, and cluster ID ***************/
  public void setOrdering(int ordering) {
    this.ordering = ordering;
    this.setKey(ordering);
  }

  public void setEncoding(double encoding) {
    this.encoding = encoding;
  }

  public void setSpecial(boolean special) {
    this.special = special;
  }

  public void setMajorClusterId(int cid) {
    this.majorClusterId = cid;
  }

  public void setMinorClusterId(int cid) {
    this.minorClusterId = cid;
  }

  public int getId() {
    return this.ID;
  }

  /////////////////////////////////////////////////////////////////////////
  //For interface comparable:

  public double getKey() {
    return this.key;
  }

  //For interface Sortable
  public void setKey(double k) {
    this.key = k;
  }

  public int compareTo(Object o) {
    double value = ( (BasicCell) o).getKey();
    if (this.getKey() < value)
      return -1;
    else if (this.getKey() > value)
      return 1;
    else
      return 0;
  }
}
